import jax
import jax.numpy as jnp
import optax

from .transform import adan


def booth(x):
    return (x[0] + 2 * x[1] - 7) ** 2 + (2 * x[0] + x[1] - 5) ** 2


def sphere(x):
    return jnp.sum(x ** 2)


def _test_optimization(func, optimizer, num_iters, initial_params, expected_optimum):
    value_grad_func = jax.value_and_grad(func)
    params = initial_params
    optimizer_state = optimizer.init(params)

    @jax.jit
    def step(params, optimizer_state):
        value, grad = value_grad_func(params)
        updates, optimizer_state = optimizer.update(grad, optimizer_state, params)
        params = optax.apply_updates(params, updates)
        return value, params, optimizer_state

    for i in range(num_iters):
        value, params, optimizer_state = step(params, optimizer_state)
        if i % 100 == 0:
            print(f'iter {i}: v={value:.4f} x={params}')

    assert jnp.allclose(params, expected_optimum, atol=1e-3)


def test_adan_sphere():
    return _test_optimization(
        func=sphere,
        optimizer=adan(learning_rate=0.01),
        num_iters=2000,
        initial_params=jnp.array([5.0, -7.5]),
        expected_optimum=jnp.array([0.0, 0.0]),
    )


def test_adan_booth():
    return _test_optimization(
        func=booth,
        optimizer=adan(learning_rate=0.01),
        num_iters=2000,
        initial_params=jnp.array([5.0, -7.5]),
        expected_optimum=jnp.array([1.0, 3.0]),
    )


def test_adam_sphere():
    return _test_optimization(
        func=sphere,
        optimizer=optax.adam(learning_rate=0.01),
        num_iters=3000,
        initial_params=jnp.array([5.0, -7.5]),
        expected_optimum=jnp.array([0.0, 0.0]),
    )


def test_adam_booth():
    return _test_optimization(
        func=booth,
        optimizer=optax.adam(learning_rate=0.01),
        num_iters=5000,
        initial_params=jnp.array([5.0, -7.5]),
        expected_optimum=jnp.array([1.0, 3.0]),
    )



import jax
import chex
import jax.numpy as jnp

from optax._src.base import GradientTransformation, Updates
from optax._src.combine import chain
from optax._src.transform import update_moment, bias_correction, scale_by_schedule
from optax._src.alias import ScalarOrSchedule, _scale_by_learning_rate
from optax._src import numerics

from typing import NamedTuple


class ScaleByAdanState(NamedTuple):
    """State for the Adan algorithm."""
    count: chex.Array
    m: Updates
    v: Updates
    n: Updates
    prev_grad: Updates


def scale_by_adan(
    b1: float = 0.02,
    b2: float = 0.08,
    b3: float = 0.01,
    eps: float = 1e-8,
) -> GradientTransformation:
    """Rescale updates according to the Adan algorithm.
    References:
      [Xie et al, 2022](https://arxiv.org/abs/2208.06677)
    Args:
      b1: the corresponding parameter from the paper.
      b2: the corresponding parameter from the paper.
      b3: the corresponding parameter from the paper.
      eps: term added to the denominator inside the square-root to improve numerical stability.
    Returns:
      A `GradientTransformation` object.
    """

    def init_fn(params):
        m = jax.tree_util.tree_map(jnp.zeros_like, params)
        v = jax.tree_util.tree_map(jnp.zeros_like, params)
        n = jax.tree_util.tree_map(jnp.zeros_like, params)
        prev_grad = jax.tree_util.tree_map(jnp.zeros_like, params)
        return ScaleByAdanState(count=jnp.zeros([], jnp.int32), m=m, v=v, n=n, prev_grad=prev_grad)

    def update_fn(updates, state, params=None):
        del params

        decay_m = 1.0 - b1
        decay_v = 1.0 - b2
        decay_n = 1.0 - b3

        m = update_moment(updates, state.m, decay=decay_m, order=1)

        grad_diff = jax.lax.cond(
            pred=state.count == 0,  # Gradient diff is zero on the first iteration
            true_fun=lambda _: jax.tree_util.tree_map(jnp.zeros_like, updates),
            false_fun=lambda _: jax.tree_util.tree_map(lambda cur, prev: cur - prev, updates, state.prev_grad),
            operand=None,
        )

        v = update_moment(grad_diff, state.v, decay=decay_v, order=1)

        second_order_moment_updates = jax.tree_util.tree_map(
            lambda cur, diff: cur + decay_v * diff, updates, grad_diff)
        n = update_moment(second_order_moment_updates, state.n, decay=decay_n, order=2)

        count_inc = numerics.safe_int32_increment(state.count)
        m_hat = bias_correction(m, decay_m, count_inc)
        v_hat = bias_correction(v, decay_v, count_inc)
        n_hat = bias_correction(n, decay_n, count_inc)

        new_updates = jax.tree_util.tree_map(
            lambda mm, vv, nn: (mm + decay_v * vv) / (jnp.sqrt(nn) + eps), m_hat, v_hat, n_hat)

        return new_updates, ScaleByAdanState(count=count_inc, m=m, v=v, n=n, prev_grad=updates)

    return GradientTransformation(init_fn, update_fn)


def adan(
    learning_rate: ScalarOrSchedule,
    b1: float = 0.02,
    b2: float = 0.08,
    b3: float = 0.01,
    eps: float = 1e-8,
) -> GradientTransformation:
    return chain(
        scale_by_adan(b1=b1, b2=b2, b3=b3, eps=eps),
        _scale_by_learning_rate(learning_rate),
    )



from setuptools import setup
from pathlib import Path

_CURRENT_DIR = Path(__file__).parent


def _get_long_description():
    return (_CURRENT_DIR / "README.md").read_text()


def _get_version():
    with open(_CURRENT_DIR / 'optax_adan' / '__init__.py', 'r') as init_file:
        for line in init_file:
            if line.startswith('__version__') and '=' in line:
                version = line[line.find('=') + 1:].strip(' \'"\n')
        if version:
            return version
        raise ValueError('`__version__` not defined in `optax_adan/__init__.py`')


setup(
    name='optax-adan',
    version=_get_version(),
    description='An implementation of adan optimization algorithm for optax.',
    long_description=_get_long_description(),
    long_description_content_type="text/markdown",
    url='https://github.com/hr0nix/optax-adan',
    author='Boris Yangel',
    author_email='boris.jangel@gmail.com',
    license='Apache License 2.0',
    packages=['optax_adan'],
    install_requires=['optax'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
